#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-ANN.
# @File         : ANN
# @Time         : 2019-12-04 20:05
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :


import faiss
import numpy as np


class ANNFaiss(object):
    """Flat支持: https://github.com/Jie-Yuan/faiss_note/blob/master/4.Faiss%20indexes%20%E8%BF%9B%E9%98%B6%E6%93%8D%E4%BD%9C.ipynb
    恢复原数据
    从index中移除向量
    搜索距离范围内的向量
    拆分/合并index

    cpu_index.make_direct_map()
    cpu_index.reconstruct(0)
    ann.index.reconstruct
    ann.index.reconstruct_c
    ann.index.reconstruct_n
    ann.index.reconstruct_n_c
    ann.index.search_and_reconstruct
    ann.index.search_and_reconstruct_c


    数据集的大小
    在高效检索的index中，聚类是其中的基础操作，数据量的大小主要影响聚类过程。

    如果小于1M， 使用"...,IVFx,..."
    N是数据集中向量个数，x一般取值[4sqrt(N),16sqrt(N)],需要30x ～ 256x个向量的数据集去训练。

    如果在1M-10M，使用"...,IMI2x10,..."
    使用k-means将训练集聚类为2^10个类，但是执行过程是在数据集的两半部分独立执行，即聚类中心有2^(2*10)个。

    如果在10M-100M，使用"...,IMI2x12,..."

    一个簇至少39个样本：一般2的N次方，我们一般用2**14或者更小一点 >> topk
    经验公式 2 ** (np.log2(n) // 2 + 2)
    """

    def __init__(self):
        self.nogpu_index_factory = {'HNSW', 'SQ'}

    def train(self, data, index_factory='Flat', metric=None, noramlize=False):
        """

        :param data:
        :param index_factory:
            https://www.cnblogs.com/houkai/p/9316172.html
            https://blog.csdn.net/xiaoxu2050/article/details/84982478
            https://github.com/facebookresearch/faiss/wiki/Faiss-indexes
            https://github.com/liqima/faiss_note/blob/master/4.Faiss%20indexes%20IO%E5%92%8Cindex%20factory.ipynb
        :param metric:
            faiss.METRIC_BrayCurtis
            faiss.METRIC_Canberra
            faiss.METRIC_INNER_PRODUCT: L2之后内积≈cosine
            faiss.METRIC_JensenShannon
            faiss.METRIC_L1
            faiss.METRIC_L2
            faiss.METRIC_Linf
            faiss.METRIC_Lp
        :return:
        """
        if noramlize:
            data = self.noramlize(data)
        assert data.dtype == 'float32', "TODO: np.array([]).astype('float32')"
        dim = data.shape[1]
        args = [dim, index_factory, metric] if metric else [dim, index_factory]

        self.index = faiss.index_factory(*args)

        if faiss.get_num_gpus() > 0:
            if any(index_factory.__contains__(i) for i in self.nogpu_index_factory):
                pass
                print(f"Donot Support GPU: {index_factory}")
            else:
                # gpu_index_flat = faiss.index_cpu_to_gpu(res, 0, index_flat)
                self.index = faiss.index_cpu_to_all_gpus(self.index)

        print(f"Train ...")
        self.index.train(data)
        self.index.add(data)

        print(f"Ntotal: {self.index.ntotal}")

    def search(self, data, topK=10, nprobe=1, k_factor=1):
        """

        :param data:
        :param topK:
        :param nprobe: nprobe参数始终是调整速度和结果精度之间权衡的一种方式。
        :param k_factor:
        :return:
        """
        self.index.k_factor = k_factor  # 搜索阶段会首    先选取 k_factor*topK，重排序
        self.index.nprobe = nprobe  # default nprobe is 1, try a few more
        return self.index.search(data, topK)

    def write_index(self, file_name="index_file.index"):
        # faiss.index_cpu_to_all_gpus(index)
        if faiss.get_num_gpus() > 0:
            index = faiss.index_gpu_to_cpu(self.index)
            faiss.write_index(index, str(file_name))
        else:
            faiss.write_index(self.index, str(file_name))

    def read_index(self, file_name="index_file.index"):
        index = faiss.read_index(str(file_name))
        # if faiss.get_num_gpus() > 0:
        #     index = faiss.index_cpu_to_gpu()
        # index_new = faiss.clone_index(index) # 复制索引
        return index

    def noramlize(self, x):
        if len(x.shape) > 1:
            return x / np.clip(x ** 2, 1e-12, None).sum(axis=1).reshape((-1, 1) + x.shape[2:]) ** 0.5
        else:
            return x / np.clip(x ** 2, 1e-12, None).sum() ** 0.5



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-ANN.
# @File         : utils
# @Time         : 2019-12-06 16:59
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
import numpy as np
import pandas as pd
from pathlib import Path
from concurrent.futures import ProcessPoolExecutor

# 并行加载
from functools import partial


def load_vector(files=Path('./ann_data/').glob('part*')):
    """

    :param files:
    :return: idx2word and vector data
    """
    with ProcessPoolExecutor(5) as pool:
        dfs = pool.map(partial(pd.read_csv, sep=' ', header=None), files)
        df = pd.concat(dfs, ignore_index=True)
    return df.iloc[:, 0].to_dict(), np.array(df.values[:, 1:].tolist(), 'float32')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-ANN.
# @File         : ANNMilvus
# @Time         : 2020/4/27 11:34 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


# https://github.com/milvus-io/docs/blob/master/site/en/guides/milvus_operation.md#createdrop-indexes-in-a-collection
# https://github.com/milvus-io/pymilvus/tree/master/examples/indexes
# https://raw.githubusercontent.com/milvus-io/pymilvus/0.7.0/examples/example.py
# TODO: 补充对应的rest api
# m.milvus.preload_table

import sys
import time
from milvus import Milvus, IndexType, MetricType


class ANNMilvus(object):

    def __init__(self, host='10.46.30.210', port='19530'):
        self.milvus = Milvus()

        print("Client Version:", self.milvus.client_version())

        status = self.milvus.connect(host, port)

        if status.OK():
            print("Server connected.")
            status, self.collections = self.milvus.show_collections()
        else:
            print("Server connect fail.")
            sys.exit(1)

        print("Server Version:", self.milvus.server_version()[-1])

    def describe_collection(self, collection_name='test'):
        print("Collections: ", self.collections)
        print(f"Describe: {self.milvus.describe_collection(collection_name)[-1]}")
        print(f"Vector number in {collection_name}: {self.milvus.count_collection(collection_name)}")

    def create_collection(self, collection_name='test', dimension=768, index_file_size=1024, metric_type=MetricType.IP,
                          **kwargs):

        status, ok = self.milvus.has_collection(collection_name)
        if ok:
            print(f"{collection_name} already exists！！！")
        else:
            params = {
                'collection_name': collection_name,
                'dimension': dimension,
                'index_file_size': index_file_size,  # optional index_file_size：文件到达这个大小的时候，milvus开始为这个文件创建索引。
                'metric_type': metric_type  # optional
            }

            self.milvus.create_collection({**params, **kwargs})

    def insert(self, collection_name, records, ids=None, partition_tag=None, params=None, timeout=-1):
        status, self.ids = self.milvus.insert(collection_name, records, ids, partition_tag, params, timeout)
        time.sleep(3)
        status, result = self.milvus.count_collection(collection_name)
        print(status)
        print("count_collection: ", result)

    def create_index(self, collection_name, index_type=IndexType.IVFLAT, index_param=None, timeout=-1):
        """drop_index

        :param collection_name:
        :param index_type:
        :param index_param:
            index_param = {
                'nlist': 2048
                }
        :param timeout:
        :return:
        """
        # You can search vectors without creating index. however, Creating index help to
        # search faster
        status = self.milvus.create_index(collection_name, index_type, index_param, timeout)
        print(status)

        # describe index, get information of index
        status, index = self.milvus.describe_index(collection_name)
        print(status)
        print(index)

    def drop_collection(self, collection_name, partition_tag=None, timeout=10):

        if partition_tag:
            status = self.milvus.drop_partition(collection_name, partition_tag, timeout)
        else:
            status = self.milvus.drop_collection(collection_name, timeout)

        print(status)
        time.sleep(1)

    def search(self, collection_name, top_k, query_records, partition_tags=None, params=None):
        """

        :param collection_name:
        :param top_k:
        :param query_records:
        :param partition_tags:
        :param params: {'nprobe': 16}
        :return:
        """
        status, results = self.milvus.search(collection_name, top_k, query_records, partition_tags, params)
        print(status)
        print(results)

    def get_vector_by_id(self):
        """
        ann.milvus.delete_by_id
        ann.milvus.get_vector_by_id
        ann.milvus.get_vector_ids
        """
        pass
